#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from oc_ocdm.abstract_entity import AbstractEntity

if TYPE_CHECKING:
    from typing import List, ClassVar, Dict, Optional
    from rdflib import URIRef, Graph


class AbstractSet(ABC):
    """
    Abstract class which represents a generic set of entities.
    It is the base class for each concrete set of entities.
    """

    short_name_to_type_iri: ClassVar[Dict[str, URIRef]] = {}

    def __init__(self) -> None:
        """
        Constructor of the ``AbstractSet`` class.
        """
        self.res_to_entity: Dict[URIRef, AbstractEntity] = {}

    def graphs(self) -> List[Graph]:
        """
        A utility method that allows to retrieve the list of ``rdflib.Graph``
        instances corresponding to each entity contained in the set.

        :return: The requested list of graphs
        """
        result: List[Graph] = []
        for entity in self.res_to_entity.values():
            if len(entity.g) > 0:
                result.append(entity.g)
        return result

    @abstractmethod
    def get_entity(self, res: URIRef) -> Optional[AbstractEntity]:
        """
        Method signature for concrete implementations that allow
        to retrieve a contained entity identified by its URI.

        :param res: The URI that identifies the requested entity
        :type res: URIRef
        :return: The requested entity if found, None otherwise
        """
        raise NotImplementedError

    @staticmethod
    def get_graph_iri(g: Graph) -> str:
        """
        A utility method that allows to retrieve the IRI which represents
        the name of a given named graph.

        **NOTE: this is a static function!**

        :param g: The named graph whose name will be returned
        :type g: Graph
        :return: The requested string whose content is the IRI associated to the given named graph
        """
        return str(g.identifier)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

import os
from shutil import copymode, move
from tempfile import mkstemp
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import BinaryIO, Tuple, List, Dict

from oc_ocdm.counter_handler.counter_handler import CounterHandler
from oc_ocdm.support.support import is_string_empty


class FilesystemCounterHandler(CounterHandler):
    """A concrete implementation of the ``CounterHandler`` interface that persistently stores
    the counter values within the filesystem."""

    _initial_line_len: int = 3
    _trailing_char: str = " "

    def __init__(self, info_dir: str, supplier_prefix: str = "") -> None:
        """
        Constructor of the ``FilesystemCounterHandler`` class.

        :param info_dir: The path to the folder that does/will contain the counter values.
        :type info_dir: str
        :raises ValueError: if ``info_dir`` is None or an empty string.
        """
        if info_dir is None or is_string_empty(info_dir):
            raise ValueError("info_dir parameter is required!")

        if info_dir[-1] != os.sep:
            info_dir += os.sep

        self.info_dir: str = info_dir
        self.supplier_prefix: str = supplier_prefix
        self.datasets_dir: str = info_dir + 'datasets' + os.sep
        self.short_names: List[str] = ["an", "ar", "be", "br", "ci", "de", "id", "pl", "ra", "re", "rp"]
        self.metadata_short_names: List[str] = ["di"]
        self.info_files: Dict[str, str] = {key: ("info_file_" + key + ".txt")
                                           for key in self.short_names}
        self.prov_files: Dict[str, str] = {key: ("prov_file_" + key + ".txt")
                                           for key in self.short_names}

    def set_counter(self, new_value: int, entity_short_name: str, prov_short_name: str = "",
                    identifier: int = 1, supplier_prefix: str = "") -> None:
        """
        It allows to set the counter value of graph and provenance entities.

        :param new_value: The new counter value to be set
        :type new_value: int
        :param entity_short_name: The short name associated either to the type of the entity itself
         or, in case of a provenance entity, to the type of the relative graph entity.
        :type entity_short_name: str
        :param prov_short_name: In case of a provenance entity, the short name associated to the type
         of the entity itself. An empty string otherwise.
        :type prov_short_name: str
        :param identifier: In case of a provenance entity, the counter value that identifies the relative
          graph entity. The integer value '1' otherwise.
        :type identifier: int
        :raises ValueError: if ``new_value`` is a negative integer or ``identifier`` is less than or equal to zero.
        :return: None
        """
        if new_value < 0:
            raise ValueError("new_value must be a non negative integer!")

        if prov_short_name == "se":
            file_path: str = self._get_prov_path(entity_short_name, supplier_prefix)
        else:
            file_path: str = self._get_info_path(entity_short_name, supplier_prefix)
        self._set_number(new_value, file_path, identifier)

    def set_counters_batch(self, updates: Dict[Tuple[str, str], Dict[int, int]], supplier_prefix: str) -> None:
        """
        Updates counters in batch for multiple files.
        `updates` is a dictionary where the key is a tuple (entity_short_name, prov_short_name)
        and the value is a dictionary of line numbers to new counter values.
        """
        for (entity_short_name, prov_short_name), file_updates in updates.items():
            file_path = self._get_prov_path(entity_short_name, supplier_prefix) if prov_short_name == "se" else self._get_info_path(entity_short_name, supplier_prefix)
            self._set_numbers(file_path, file_updates)

    def _set_numbers(self, file_path: str, updates: Dict[int, int]) -> None:
        """
        Apply multiple counter updates to a single file.
        `updates` is a dictionary where the key is the line number (identifier)
        and the value is the new counter value.
        """
        self.__initialize_file_if_not_existing(file_path)
        with open(file_path, 'r') as file:
            lines = file.readlines()
        max_line_number = max(updates.keys())

        # Ensure the lines list is long enough
        while len(lines) < max_line_number + 1:
            lines.append("\n")  # Default counter value

        # Apply updates
        for line_number, new_value in updates.items():
            lines[line_number-1] = str(new_value).rstrip() + "\n"

        # Write updated lines back to file
        with open(file_path, 'w') as file:
            file.writelines(lines)

    def read_counter(self, entity_short_name: str, prov_short_name: str = "", identifier: int = 1, supplier_prefix: str = "") -> int:
        """
        It allows to read the counter value of graph and provenance entities.

        :param entity_short_name: The short name associated either to the type of the entity itself
         or, in case of a provenance entity, to the type of the relative graph entity.
        :type entity_short_name: str
        :param prov_short_name: In case of a provenance entity, the short name associated to the type
         of the entity itself. An empty string otherwise.
        :type prov_short_name: str
        :param identifier: In case of a provenance entity, the counter value that identifies the relative
          graph entity. The integer value '1' otherwise.
        :type identifier: int
        :raises ValueError: if ``identifier`` is less than or equal to zero.
        :return: The requested counter value.
        """
        if prov_short_name == "se":
            file_path: str = self._get_prov_path(entity_short_name, supplier_prefix)
        else:
            file_path: str = self._get_info_path(entity_short_name, supplier_prefix)
        return self._read_number(file_path, identifier)

    def increment_counter(self, entity_short_name: str, prov_short_name: str = "", identifier: int = 1, supplier_prefix: str = "") -> int:
        """
        It allows to increment the counter value of graph and provenance entities by one unit.

        :param entity_short_name: The short name associated either to the type of the entity itself
         or, in case of a provenance entity, to the type of the relative graph entity.
        :type entity_short_name: str
        :param prov_short_name: In case of a provenance entity, the short name associated to the type
         of the entity itself. An empty string otherwise.
        :type prov_short_name: str
        :param identifier: In case of a provenance entity, the counter value that identifies the relative
          graph entity. The integer value '1' otherwise.
        :type identifier: int
        :raises ValueError: if ``identifier`` is less than or equal to zero.
        :return: The newly-updated (already incremented) counter value.
        """
        if prov_short_name == "se":
            file_path: str = self._get_prov_path(entity_short_name, supplier_prefix)
        else:
            file_path: str = self._get_info_path(entity_short_name, supplier_prefix)
        return self._add_number(file_path, identifier)

    def _get_info_path(self, short_name: str, supplier_prefix: str) -> str:
        supplier_prefix = "" if supplier_prefix is None else supplier_prefix
        directory = self.info_dir if supplier_prefix == self.supplier_prefix or not self.supplier_prefix else self.info_dir.replace(self.supplier_prefix, supplier_prefix, 1)
        return directory + self.info_files[short_name]

    def _get_prov_path(self, short_name: str, supplier_prefix: str) -> str:
        supplier_prefix = "" if supplier_prefix is None else supplier_prefix
        directory = self.info_dir if supplier_prefix == self.supplier_prefix or not self.supplier_prefix else self.info_dir.replace(self.supplier_prefix, supplier_prefix, 1)
        return directory + self.prov_files[short_name]

    def _get_metadata_path(self, short_name: str, dataset_name: str) -> str:
        return self.datasets_dir + dataset_name + os.sep + 'metadata_' + short_name + '.txt'

    def __initialize_file_if_not_existing(self, file_path: str):
        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path))

        if not os.path.isfile(file_path):
            with open(file_path, 'w') as file:
                file.write("\n")

    def _read_number(self, file_path: str, line_number: int) -> int:
        if line_number <= 0:
            raise ValueError("line_number must be a positive non-zero integer number!")

        self.__initialize_file_if_not_existing(file_path)

        cur_number: int = 0
        try:
            with open(file_path, 'r') as file:
                for i, line in enumerate(file, 1):
                    if i == line_number:
                        line = line.strip()
                        if line:
                            cur_number = int(line)
                        break
                else:
                    print(file_path)
        except ValueError as e:
            print(f"ValueError: {e}")
            cur_number = 0
        except Exception as e:
            print(f"Unexpected error: {e}")
        return cur_number

    def _add_number(self, file_path: str, line_number: int = 1) -> int:
        if line_number <= 0:
            raise ValueError("line_number must be a positive non-zero integer number!")

        self.__initialize_file_if_not_existing(file_path)

        current_value = self._read_number(file_path, line_number)
        new_value = current_value + 1
        self._set_number(new_value, file_path, line_number)
        return new_value

    def _set_number(self, new_value: int, file_path: str, line_number: int = 1) -> None:
        if new_value < 0:
            raise ValueError("new_value must be a non negative integer!")

        if line_number <= 0:
            raise ValueError("line_number must be a positive non-zero integer number!")

        self.__initialize_file_if_not_existing(file_path)

        lines = []
        with open(file_path, 'r') as file:
            lines = file.readlines()

        # Ensure the file has enough lines
        while len(lines) < line_number:
            lines.append("\n")

        # Update the specific line
        lines[line_number - 1] = f"{new_value}\n"

        # Write back to the file
        with open(file_path, 'w') as file:
            file.writelines(lines)


    def set_metadata_counter(self, new_value: int, entity_short_name: str, dataset_name: str) -> None:
        """
        It allows to set the counter value of metadata entities.

        :param new_value: The new counter value to be set
        :type new_value: int
        :param entity_short_name: The short name associated either to the type of the entity itself.
        :type entity_short_name: str
        :param dataset_name: In case of a ``Dataset``, its name. Otherwise, the name of the relative dataset.
        :type dataset_name: str
        :raises ValueError: if ``new_value`` is a negative integer, ``dataset_name`` is None or
          ``entity_short_name`` is not a known metadata short name.
        :return: None
        """
        if new_value < 0:
            raise ValueError("new_value must be a non negative integer!")

        if dataset_name is None:
            raise ValueError("dataset_name must be provided!")

        if entity_short_name not in self.metadata_short_names:
            raise ValueError("entity_short_name is not a known metadata short name!")

        file_path: str = self._get_metadata_path(entity_short_name, dataset_name)
        return self._set_number(new_value, file_path, 1)

    def read_metadata_counter(self, entity_short_name: str, dataset_name: str) -> int:
        """
        It allows to read the counter value of metadata entities.

        :param entity_short_name: The short name associated either to the type of the entity itself.
        :type entity_short_name: str
        :param dataset_name: In case of a ``Dataset``, its name. Otherwise, the name of the relative dataset.
        :type dataset_name: str
        :raises ValueError: if ``dataset_name`` is None or ``entity_short_name`` is not a known metadata short name.
        :return: The requested counter value.
        """
        if dataset_name is None:
            raise ValueError("dataset_name must be provided!")

        if entity_short_name not in self.metadata_short_names:
            raise ValueError("entity_short_name is not a known metadata short name!")

        file_path: str = self._get_metadata_path(entity_short_name, dataset_name)
        return self._read_number(file_path, 1)

    def increment_metadata_counter(self, entity_short_name: str, dataset_name: str) -> int:
        """
        It allows to increment the counter value of metadata entities by one unit.

        :param entity_short_name: The short name associated either to the type of the entity itself.
        :type entity_short_name: str
        :param dataset_name: In case of a ``Dataset``, its name. Otherwise, the name of the relative dataset.
        :type dataset_name: str
        :raises ValueError: if ``dataset_name`` is None or ``entity_short_name`` is not a known metadata short name.
        :return: The newly-updated (already incremented) counter value.
        """
        if dataset_name is None:
            raise ValueError("dataset_name must be provided!")

        if entity_short_name not in self.metadata_short_names:
            raise ValueError("entity_short_name is not a known metadata short name!")

        file_path: str = self._get_metadata_path(entity_short_name, dataset_name)
        return self._add_number(file_path, 1)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import Optional
    from rdflib import URIRef
    from oc_ocdm.graph.entities.bibliographic.responsible_agent import ResponsibleAgent
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class AgentRole(BibliographicEntity):
    """Agent role (short: ar): a particular role held by an agent with respect to a bibliographic resource."""

    @accepts_only('ar')
    def merge(self, other: AgentRole) -> None:
        """
        The merge operation allows combining two ``AgentRole`` entities into a single one,
        by marking the second entity as to be deleted while also copying its data into the current
        ``AgentRole``. Moreover, every triple from the containing ``GraphSet`` referring to the second
        entity gets "redirected" to the current entity: **every other reference contained inside a
        different source (e.g. a triplestore) must be manually handled by the user!**

        In case of functional properties, values from the current entity get overwritten
        by those coming from the second entity while, in all other cases, values from the
        second entity are simply appended to those of the current entity. In this context,
        ``rdfs:label`` is considered as a functional property, while ``rdf:type`` is not.

        :param other: The entity which will be marked as to be deleted and whose properties will
         be merged into the current entity.
        :type other: AgentRole
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        super(AgentRole, self).merge(other)

        next_ar: Optional[AgentRole] = other.get_next()
        if next_ar is not None:
            self.has_next(next_ar)

        resp_agent: Optional[ResponsibleAgent] = other.get_is_held_by()
        if resp_agent is not None:
            self.is_held_by(resp_agent)

        role_type: Optional[URIRef] = other.get_role_type()
        if role_type is not None:
            if role_type == GraphEntity.iri_publisher:
                self.create_publisher()
            elif role_type == GraphEntity.iri_author:
                self.create_author()
            elif role_type == GraphEntity.iri_editor:
                self.create_editor()

    # HAS NEXT (AgentRole)
    def get_next(self) -> Optional[AgentRole]:
        """
        Getter method corresponding to the ``oco:hasNext`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_has_next, 'ar')
        if uri is not None:
            return self.g_set.add_ar(self.resp_agent, self.source, uri)

    @accepts_only('ar')
    def has_next(self, ar_res: AgentRole) -> None:
        """
        Setter method corresponding to the ``oco:hasNext`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The previous role in a sequence of agent roles of the same type associated with the
        same bibliographic resource (so as to define, for instance, an ordered list of authors).`

        :param ar_res: The value that will be set as the object of the property related to this method
        :type ar_res: AgentRole
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_next()
        self.g.add((self.res, GraphEntity.iri_has_next, ar_res.res))

    def remove_next(self) -> None:
        """
        Remover method corresponding to the ``oco:hasNext`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_next, None))

    # IS HELD BY (ResponsibleAgent)
    def get_is_held_by(self) -> Optional[ResponsibleAgent]:
        """
        Getter method corresponding to the ``pro:isHeldBy`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_is_held_by, 'ra')
        if uri is not None:
            return self.g_set.add_ra(self.resp_agent, self.source, uri)

    @accepts_only('ra')
    def is_held_by(self, ra_res: ResponsibleAgent):
        """
        Setter method corresponding to the ``pro:isHeldBy`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The agent holding this role with respect to a particular bibliographic resource.`

        :param ra_res: The value that will be set as the object of the property related to this method
        :type ra_res: ResponsibleAgent
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_is_held_by()
        self.g.add((self.res, GraphEntity.iri_is_held_by, ra_res.res))

    def remove_is_held_by(self) -> None:
        """
        Remover method corresponding to the ``pro:isHeldBy`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_is_held_by, None))

    # HAS ROLE TYPE
    def get_role_type(self) -> Optional[URIRef]:
        """
        Getter method corresponding to the ``pro:withRole`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_with_role)
        return uri

    def create_publisher(self) -> None:
        """
        Setter method corresponding to the ``pro:withRole`` RDF predicate.
        It implicitly sets the object value ``pro:publisher``.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The specific type of role under consideration (e.g. author, editor or publisher).`

        :return: None
        """
        self.remove_role_type()
        self.g.add((self.res, GraphEntity.iri_with_role, GraphEntity.iri_publisher))

    def create_author(self) -> None:
        """
        Setter method corresponding to the ``pro:withRole`` RDF predicate.
        It implicitly sets the object value ``pro:author``.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The specific type of role under consideration (e.g. author, editor or publisher).`

        :return: None
        """
        self.remove_role_type()
        self.g.add((self.res, GraphEntity.iri_with_role, GraphEntity.iri_author))

    def create_editor(self) -> None:
        """
        Setter method corresponding to the ``pro:withRole`` RDF predicate.
        It implicitly sets the object value ``pro:editor``.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The specific type of role under consideration (e.g. author, editor or publisher).`

        :return: None
        """
        self.remove_role_type()
        self.g.add((self.res, GraphEntity.iri_with_role, GraphEntity.iri_editor))

    def remove_role_type(self) -> None:
        """
        Remover method corresponding to the ``pro:withRole`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_with_role, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from rdflib import Graph, Namespace, URIRef

from oc_ocdm.abstract_entity import AbstractEntity
from oc_ocdm.graph.graph_entity import GraphEntity

if TYPE_CHECKING:
    from typing import ClassVar, Dict

    from oc_ocdm.prov.prov_set import ProvSet


class ProvEntity(AbstractEntity):
    """Snapshot of entity metadata (short: se): a particular snapshot recording the
    metadata associated with an individual entity (either a bibliographic entity or an
    identifier) at a particular date and time, including the agent, such as a person,
    organisation or automated process that created or modified the entity metadata.
    """

    PROV: ClassVar[Namespace] = Namespace("http://www.w3.org/ns/prov#")

    iri_entity: ClassVar[URIRef] = PROV.Entity
    iri_generated_at_time: ClassVar[URIRef] = PROV.generatedAtTime
    iri_invalidated_at_time: ClassVar[URIRef] = PROV.invalidatedAtTime
    iri_specialization_of: ClassVar[URIRef] = PROV.specializationOf
    iri_was_derived_from: ClassVar[URIRef] = PROV.wasDerivedFrom
    iri_had_primary_source: ClassVar[URIRef] = PROV.hadPrimarySource
    iri_was_attributed_to: ClassVar[URIRef] = PROV.wasAttributedTo
    iri_description: ClassVar[URIRef] = GraphEntity.DCTERMS.description
    iri_has_update_query: ClassVar[URIRef] = GraphEntity.OCO.hasUpdateQuery

    short_name_to_type_iri: ClassVar[Dict[str, URIRef]] = {
        'se': iri_entity
    }

    def __init__(self, prov_subject: GraphEntity, g: Graph, p_set: ProvSet,
                 res: URIRef = None, resp_agent: str = None, source: str = None,
                 res_type: URIRef = None, count: str = None, label: str = None,
                 short_name: str = "") -> None:
        super(ProvEntity, self).__init__()
        self.prov_subject: GraphEntity = prov_subject

        self.g: Graph = g
        self.resp_agent: str = resp_agent
        self.source: str = source
        self.short_name: str = short_name
        self.p_set: ProvSet = p_set

        # If res was not specified, create from scratch the URI reference for this entity,
        # otherwise use the provided one
        if res is None:
            self.res = self._generate_new_res(g, count, short_name)
        else:
            self.res = res

        if p_set is not None:
            # If not already done, register this ProvEntity instance inside the ProvSet
            if self.res not in p_set.res_to_entity:
                p_set.res_to_entity[self.res] = self

        # Add mandatory information to the entity graph
        self._create_type(res_type)
        if label is not None:
            self.create_label(label)

    @staticmethod
    def _generate_new_res(g: Graph, count: str, short_name: str) -> URIRef:
        return URIRef(str(g.identifier) + short_name + "/" + count)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Tuple
    from rdflib import URIRef
    from rdflib.compare import IsomorphicGraph
    from oc_ocdm.abstract_entity import AbstractEntity

from rdflib import Graph
from rdflib.compare import graph_diff, to_isomorphic


def get_delete_query(graph_iri: URIRef, data: Graph) -> Tuple[str, int]:
    num_of_statements: int = len(data)
    if num_of_statements <= 0:
        return "", 0
    else:
        statements: str = data.serialize(format="nt11").replace('\n', '')
        delete_string: str = f"DELETE DATA {{ GRAPH <{graph_iri}> {{ {statements} }} }}"
        return delete_string, num_of_statements


def get_insert_query(graph_iri: URIRef, data: Graph) -> Tuple[str, int]:
    num_of_statements: int = len(data)
    if num_of_statements <= 0:
        return "", 0
    else:
        statements: str = data.serialize(format="nt11").replace('\n', '')
        insert_string: str = f"INSERT DATA {{ GRAPH <{graph_iri}> {{ {statements} }} }}"
        return insert_string, num_of_statements


def get_update_query(entity: AbstractEntity, entity_type: str = "graph") -> Tuple[str, int, int]:
    if entity_type in ["graph", "metadata"]:
        to_be_deleted: bool = entity.to_be_deleted
        preexisting_graph: Graph = entity.preexisting_graph
    elif entity_type == "prov":
        to_be_deleted: bool = False
        preexisting_graph: Graph = Graph(identifier=entity.g.identifier)

    if to_be_deleted:
        delete_string, removed_triples = get_delete_query(entity.g.identifier, preexisting_graph)
        if delete_string != "":
            return delete_string, 0, removed_triples
        else:
            return "", 0, 0
    else:
        preexisting_iso: IsomorphicGraph = to_isomorphic(preexisting_graph)
        current_iso: IsomorphicGraph = to_isomorphic(entity.g)
        if preexisting_iso == current_iso:
            # Both graphs have exactly the same content!
            return "", 0, 0
        in_both, in_first, in_second = graph_diff(preexisting_iso, current_iso)
        delete_string, removed_triples = get_delete_query(entity.g.identifier, in_first)
        insert_string, added_triples = get_insert_query(entity.g.identifier, in_second)
        if delete_string != "" and insert_string != "":
            return delete_string + '; ' + insert_string, added_triples, removed_triples
        elif delete_string != "":
            return delete_string, 0, removed_triples
        elif insert_string != "":
            return insert_string, added_triples, 0
        else:
            return "", 0, 0